from pyspark.sql.types import StringType, FloatType, StructType, StructField, Row

"""
"
" default NLU output formatter which parse keywords, concepts, sentiment and emotion
"
"""
def default_features_parser(response, **params):
    """
    @param::output: the output json object from STT
    @return:the transcript join by period in string format
    """
    data = {}
    features = params["features"]
    keywords_limit = features.keywords.limit
    concepts_limit = features.concepts.limit

    # extract keyword data
    if "keywords" in response:
        for i in range(keywords_limit):
            if i < len(response["keywords"]):
                kw = response["keywords"][i]
                data['keyword_%d'%(i)] = kw['text']
                data['keyword_%d_score'%(i)] = kw['relevance']
            else: # when less keywords extracted than limit
                data['keyword_%d'%(i)] = None
                data['keyword_%d_score'%(i)] = None
    # extract concept
    if "concepts" in response:
        for i in range(concepts_limit):
            if i < len(response["concepts"]):
                concept = response["concepts"][i]
                data['concept_%d'%(i)] = concept['text']
                data['concept_%d_score'%(i)] = concept['relevance']
            else: # when less concept extracted than limit
                data['concept_%d'%(i)] = None
                data['concept_%d_score'%(i)] = None
    # extract sentiment
    if "sentiment" in response:
        data["sentiment_score"] = response["sentiment"]["document"]["score"]
        data["sentiment_label"] = response["sentiment"]["document"]["label"]
    # extract "emotion"
    if "emotion" in response:
        data["sadness_score"] = response["emotion"]["document"]["emotion"]["sadness"]
        data["joy_score"] = response["emotion"]["document"]["emotion"]["joy"]
        data["fear_score"] = response["emotion"]["document"]["emotion"]["fear"]
        data["disgust_score"] = response["emotion"]["document"]["emotion"]["disgust"]
        data["anger_score"] = response["emotion"]["document"]["emotion"]["anger"]
     
    return Row(**data)


"""
"
" pyspark transformer consume IBM Watson service
" @ref: https://stackoverflow.com/questions/41399399/serialize-a-custom-transformer-using-python-to-be-used-within-a-pyspark-ml-pipel/52467470#52467470
" @ref: https://stackoverflow.com/questions/32331848/create-a-custom-transformer-in-pyspark-ml
"
"""

from pyspark import keyword_only
from pyspark.sql import functions as F
from pyspark.sql.types import StringType, FloatType, StructType, StructField, Row
from pyspark.sql import DataFrame
from pyspark.ml.pipeline import Transformer
from pyspark.ml.param.shared import HasInputCol, HasOutputCol, Param
from pyspark.ml.util import DefaultParamsReadable, DefaultParamsWritable


class WatsonServiceTransformer(Transformer, 
                               HasInputCol, 
                               HasOutputCol,
                               DefaultParamsReadable,
                               DefaultParamsWritable):
    
    """
    "
    " set init transformer and set parameters
    "
    """
    @keyword_only
    def __init__(self, 
                 inputCol=None, 
                 outputCol=None, 
                 service=None):
        """
        @param::inputCol: the input column name contains sound file name
        @param::outputCol: the output column name
        @param::service: the IBM service object
        @return: none
        """
        super(WatsonServiceTransformer, self).__init__()
        self.service = Param(self, "service", None)
        self._setDefault(service=None)
        kwargs = self._input_kwargs
        self._set(**kwargs)
        # make sure parameter: token, endpoint set properly.
        if self.service == None:
            raise ValueError('> IBM service configuration must be provided.')
    
    """
    "
    " set parameters, called from self._set(), inherited method.
    "
    """
    @keyword_only
    def setParams(self, inputCol=None, outputCol=None, stt=None):
        """
        @param::inputCol: the input column name contains sound file name
        @param::outputCol: the output column name
        @param::service:: the IBM service object
        @return: none
        """
        kwargs = self._input_kwargs
        return self._set(**kwargs)
    
    """
    "
    " set IBM STT configuration includes: api token/endpoint, params to IBM STT service
    "
    """
    def setService(self, value):
        """
        @param::value: the IBM service API object
        @return: None
        """
        return self._set(service=value)
    
    """
    "
    " get the IBM STT configuration object
    "
    """
    def getService(self):
        """
        @param::value: the IBM service API object
        @return: None
        """
        return self.getOrDefault(self.service)

    """
    "
    " set input columns name
    "
    """
    def setInputCol(self, value):
        """
        @param::value: the input columns name contains the sound file name
        @return: None
        """
        return self._set(inputCol=value)

    """
    "
    " set output columns name
    "
    """
    def setOutputCol(self, value):
        """
        @param::value: the output column name contains the output result
        @return: None
        """
        return self._set(outputCol=value)

    
    """
    " perform the transform using provided IBM service api
    """
    def _transform(self, df:DataFrame) -> DataFrame:
        """
        @param::df: the pyspark dataframe
        @return: the transformed dataframe
        """
        stt_udf = F.udf(self.getService()(), self.getService()().get_return_type())
        df = df.withColumn(self.getOutputCol(), stt_udf(F.col(self.getInputCol())))
        return df


"""
"
" IBM NLU Service Executable Class
" docu: https://cloud.ibm.com/apidocs/natural-language-understanding?code=python#analyze-text
"
"""

import json
from pyspark import keyword_only
from ibm_watson import NaturalLanguageUnderstandingV1
from ibm_cloud_sdk_core.authenticators import IAMAuthenticator
from ibm_watson.natural_language_understanding_v1 import Features, KeywordsOptions, ConceptsOptions, SentimentOptions, EmotionOptions

class NLU():
    
    @keyword_only
    def __init__(self, token, endpoint, parser, return_type, **params):
        """
        @param::token: the IBM NLU API access token
        @param::endpoint: the endpoint url for the NLU API
        @param::paser: the object parse NLU API response
        @param::return_type: executable object return the return_type of NLU __call__
        @param::params: the kv params passing to underlying NaturalLanguageUnderstandingV1 constructor
        @return: the output parsed by parser object
        """
        self.token = token
        self.endpoint = endpoint
        self.parser = parser
        self.params = params
        self.return_type = return_type
        
    def __call__(self, text):
        """
        @param::text: the text to perform NLU
        @return: the output formatted by formatter object
        """

        # init nlu client
        authenticator = IAMAuthenticator(self.token)
        nlu = NaturalLanguageUnderstandingV1( version='2019-07-12',authenticator=authenticator)
        nlu.set_service_url(self.endpoint)

        response = nlu.analyze(text = text,
                               **self.params).get_result()
        return self.parser(response, **self.params)
    
    def get_return_type(self):
        return_type = self.return_type(**self.params)
        return return_type



import json

"""
"
" default STT output interpreter which assume one alternative transcript and no speaker detection
"
"""
def text_parser(output, **params):
    """
    @param::output: the output json object from STT
    @return:the transcript join by period in string format
    """
    transcripts = [doc['alternatives'][0]['transcript'].strip() for doc in output['results']]
    return '. '.join(transcripts) + '.'


from pyspark.sql.types import StringType, FloatType, StructType, StructField, Row


"""
"
" return the default NLU return type
"
"""
def nlu_default_return_type(**params):
    """
    @param::num_keywords: the number of keywords extracted by NLU
    @param::num_concpets: the number of concepts extracted by NLU
    @return: the defined return type
    """
    fields = []
    features = params["features"]
    num_keywords = features.keywords.limit
    num_concepts = features.concepts.limit
    # populate keyword fields
    for i in range(num_keywords):
        fields.append(StructField("keyword_%d"%(i), StringType(), True))
        fields.append(StructField("keyword_%d_score"%(i), FloatType(), True))

    # populate concpet fields
    for i in range(num_concepts):
        fields.append(StructField("concept_%d"%(i), StringType(), True))
        fields.append(StructField("concept_%d_score"%(i), FloatType(), True))

    # populate other fields
    fields.extend([StructField("sentiment_score", FloatType(), True),
                   StructField("sentiment_label", StringType(), True),
                   StructField("sadness_score", FloatType(), True),
                   StructField("joy_score", FloatType(), True),
                   StructField("fear_score", FloatType(), True),
                   StructField("disgust_score", FloatType(), True),
                   StructField("anger_score", FloatType(), True)])

    return StructType(fields)


"""
"
" return the default STT return type
"
"""
def stt_default_return_type(**params):
    """
    @return: the defined return type
    """
    return StringType()


"""
"
" IBM STT Service Executable Class
" docu: https://cloud.ibm.com/apidocs/speech-to-text
"
"""

import json
from pyspark import keyword_only
from ibm_watson import SpeechToTextV1
from ibm_cloud_sdk_core.authenticators import IAMAuthenticator

class STT():
    
    @keyword_only
    def __init__(self, token, endpoint, reader, parser, return_type, **params):
        """
        @param::token: the IBM STT API access token
        @param::endpoint: the endpoint url for the STT API
        @param::reader: the object read audio stream using audio file name/id
        @param::paser: the object parse STT API response
        @param::return_type: the return type of STT __call__
        @param::params: the kv params passing to underlying SpeechToTextV1 constructor
        @return: the output formatted by formatter executable
        """
        self.token = token
        self.endpoint = endpoint
        self.reader = reader
        self.parser = parser
        self.params = params
        self.return_type = return_type
    
    def __call__(self, audio_file):
        """
        @param::audio_file: the audio filename/id for reader to retrieve the audio stream
        @return: the output formatted by formatter object
        """
        # load asset
        audio_stream = self.reader(audio_file)

        # init stt client
        authenticator = IAMAuthenticator(self.token)
        stt = SpeechToTextV1(authenticator=authenticator)
        stt.set_service_url(self.endpoint)
        response = stt.recognize(
            audio=audio_stream,
            **self.params
        ).get_result()
        return self.parser(response, **self.params)
    
    def get_return_type(self):
        return_type = self.return_type(**self.params)
        return return_type